package com.example.survey.util;

import org.apache.commons.io.FileUtils;
import org.apache.poi.ooxml.POIXMLDocument;
import org.apache.poi.openxml4j.opc.OPCPackage;
import org.apache.poi.xwpf.usermodel.Document;
import org.apache.poi.xwpf.usermodel.PositionInParagraph;
import org.apache.poi.xwpf.usermodel.TextSegment;
import org.apache.poi.xwpf.usermodel.XWPFParagraph;
import org.apache.poi.xwpf.usermodel.XWPFRun;
import org.apache.poi.xwpf.usermodel.XWPFTable;
import org.apache.poi.xwpf.usermodel.XWPFTableCell;
import org.apache.poi.xwpf.usermodel.XWPFTableRow;
import org.apache.poi.xwpf.usermodel.XWPFRun.FontCharRange;
import org.apache.xmlbeans.XmlCursor;
import org.graalvm.compiler.word.Word;
import org.springframework.stereotype.Component;

import javax.lang.model.util.ElementScanner6;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.Map.Entry;
import java.util.regex.Pattern;

/**
 * @author Pope
 */
// @Component
public class WordUtil {

    private CustomXWPFDocument document;
    private HashMap<String, Method> key2methods;
    private Map<String, Object> values;
    private String outputPath;
    private File temp;

    public WordUtil(Map<String, Object> values, String templatePath, String outputPath)
            throws FileNotFoundException, IOException, NoSuchMethodException, SecurityException {
        temp = new File(UUID.randomUUID() + "t.docx");
        this.outputPath = outputPath;
        FileUtils.copyFile(new File(templatePath), temp);
        OPCPackage pack = POIXMLDocument.openPackage(temp.getAbsolutePath());
        document = new CustomXWPFDocument(pack);
        this.values = values;

        key2methods = new HashMap<>();

        key2methods.put("$", getClass().getDeclaredMethod("processText", XWPFTableCell.class, String.class));
        key2methods.put("#", getClass().getDeclaredMethod("processImage", XWPFTableCell.class, String.class));
        key2methods.put("@", getClass().getDeclaredMethod("processTable", XWPFTableCell.class, String.class));
    }

    public void export2word() throws FileNotFoundException, IOException, NoSuchMethodException, SecurityException {

        List<XWPFTable> tables = document.getTables();
        HashMap<String, XWPFTableCell> runs = new HashMap<>();
        for (XWPFTable table : tables) {
            // 遍历表格
            List<XWPFTableRow> rows = table.getRows();
            for (int i = 0; i < rows.size(); i++) {
                XWPFTableRow row = table.getRow(i);
                List<XWPFTableCell> cells = row.getTableCells();
                for (int j = 0; j < cells.size(); j++) {
                    XWPFTableCell cell = cells.get(j);
                    String key = cell.getText();
                    if (key.length() == 0) {
                        continue;
                    }
                    String keyHead = key.substring(0, 1);
                    String trueKey = key.substring(1);
                    if (key2methods.containsKey(keyHead)) {
                        cell.removeParagraph(0);
                    } else {
                        continue;
                    }
                    if (values.containsKey(trueKey)) {
                        runs.put(key, cell);
                    } else {
                        cell.setText("");
                    }
                }
            }
        }

        for (Entry<String, XWPFTableCell> entry : runs.entrySet()) {
            String key = entry.getKey();
            XWPFTableCell cell = entry.getValue();
            Method method = key2methods.get(key.subSequence(0, 1));
            String trueKey = key.substring(1);
            try {
                method.invoke(this, cell, values.get(trueKey));
                // System.out.println(key + ":" + values.get(key));
            } catch (Exception e) {
                e.printStackTrace();
                cell.setText("");
            }
        }
        OutputStream outputStream = new FileOutputStream(outputPath);
        // document.
        document.write(outputStream);
        outputStream.close();
        document.close();
        FileUtils.forceDelete(temp);
    }

    public void export2QuestionnaireWord()
            throws FileNotFoundException, IOException, NoSuchMethodException, SecurityException {
        List<XWPFParagraph> paragraphs = document.getParagraphs();
        for (XWPFParagraph paragraph : paragraphs) {
            String paraText = paragraph.getParagraphText();
            int startIndex = paraText.indexOf("&{");
            while (startIndex != -1) {
                int endIndex = paraText.indexOf("}", startIndex + 2);
                String trueKey = paraText.substring(startIndex + 2, endIndex);
                String key = "&{" + trueKey + "}";
                if (values.containsKey(trueKey)) {
                    Object valueObj = values.get(trueKey);
                    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");
                    if (valueObj instanceof ArrayList<?>) {
                        for (Object o : (List<?>) valueObj) {
                            String value = o.toString();
                            try {
                                Date date = dateFormat.parse(value);
                                replaceInParagragh(paragraph, key, new SimpleDateFormat("yyyy年MM月dd日").format(date));
                            } catch (Exception e) {
                                replaceInParagragh(paragraph, key, value);
                            }
                            break;
                        }
                    } else if (valueObj.getClass().isArray()) {
                        Object[] arr = (Object[]) valueObj;
                        try {
                            Date date = dateFormat.parse(arr[0].toString());
                            replaceInParagragh(paragraph, key, new SimpleDateFormat("yyyy年MM月dd日").format(date));
                        } catch (Exception e) {
                            replaceInParagragh(paragraph, key, arr[0].toString());
                        }
                    } else {
                        String value = valueObj.toString();
                        try {
                            Date date = dateFormat.parse(value);
                            replaceInParagragh(paragraph, key, new SimpleDateFormat("yyyy年MM月dd日").format(date));
                        } catch (Exception e) {
                            replaceInParagragh(paragraph, key, value);
                        }
                    }
                    // replaceInParagragh(paragraph, key, value);
                } else {
                    replaceInParagragh(paragraph, key, "         ");
                }
                startIndex = paraText.indexOf("&{", endIndex + 1);
            }

            startIndex = paraText.indexOf("*{");
            while (startIndex != -1) {
                int endIndex = paraText.indexOf("}", startIndex + 2);
                String trueKey = paraText.substring(startIndex + 2, endIndex);
                String key = "*{" + trueKey + "}";
                if (values.containsKey(trueKey)) {
                    TextSegment textSeg = paragraph.searchText(key, new PositionInParagraph());
                    int startPosRun = textSeg.getBeginRun();
                    Object valueObj = values.get(trueKey);
                    if (valueObj instanceof ArrayList<?>) {
                        for (Object o : (List<?>) valueObj) {
                            String oldText = o.toString() + "□";
                            TextSegment tSegment = paragraph.searchText(oldText,
                                    new PositionInParagraph(startPosRun, 0, 0));
                            if (tSegment != null) {
                                String newText = o.toString();
                                replaceInParagragh(paragraph, tSegment, oldText, newText);
                            } else {
                                oldText = "□" + o.toString();
                                tSegment = paragraph.searchText(oldText, new PositionInParagraph(startPosRun, 0, 0));
                                if (tSegment != null) {
                                    String newText = o.toString();
                                    replaceInParagragh(paragraph, tSegment, oldText, newText);
                                }
                            }
                        }
                    } else if (valueObj.getClass().isArray()) {
                        Object[] arr = (Object[]) valueObj;
                        for (int i = 0; i < arr.length; i++) {
                            String oldText = arr[i] + "□";
                            TextSegment tSegment = paragraph.searchText(oldText,
                                    new PositionInParagraph(startPosRun, 0, 0));
                            if (tSegment != null) {
                                String newText = arr[i].toString();
                                replaceInParagragh(paragraph, tSegment, oldText, newText);
                            } else {
                                oldText = "□" + arr[i];
                                tSegment = paragraph.searchText(oldText, new PositionInParagraph(startPosRun, 0, 0));
                                if (tSegment != null) {
                                    String newText = arr[i].toString();
                                    replaceInParagragh(paragraph, tSegment, oldText, newText);
                                }
                            }
                        }
                    } else {
                        String oldText = valueObj.toString() + "□";
                        TextSegment tSegment = paragraph.searchText(oldText,
                                new PositionInParagraph(startPosRun, 0, 0));
                        if (tSegment != null) {
                            String newText = valueObj.toString();
                            replaceInParagragh(paragraph, tSegment, oldText, newText);
                        } else {
                            oldText = "□" + valueObj.toString();
                            tSegment = paragraph.searchText(oldText, new PositionInParagraph(startPosRun, 0, 0));
                            if (tSegment != null) {
                                String newText = valueObj.toString();
                                replaceInParagragh(paragraph, tSegment, oldText, newText);
                            }
                        }
                    }
                }
                startIndex = paraText.indexOf("*{", endIndex + 1);

                // replaceInParagragh(paragraph, key, trueKey);

                if (trueKey.lastIndexOf("_") != -1) {
                    Pattern pattern = Pattern.compile("-?[0-9]+(\\.[0-9]+)?");
                    if (trueKey.substring(trueKey.lastIndexOf("_") + 1).length() > 0
                            && pattern.matcher(trueKey.substring(trueKey.lastIndexOf("_") + 1)).matches()) {
                        replaceInParagragh(paragraph, key, trueKey.substring(0, trueKey.lastIndexOf("_")));
                    }
                } else {
                    switch (trueKey) {
                        case "单位级别":
                            replaceInParagragh(paragraph, key, "");
                            break;
                        default:
                            replaceInParagragh(paragraph, key, trueKey);
                            break;
                    }
                }
            }
        }

        List<XWPFTable> tables = document.getTables();
        HashMap<String, XWPFTableCell> runs = new HashMap<>();
        for (XWPFTable table : tables) {
            // 遍历表格
            List<XWPFTableRow> rows = table.getRows();
            for (int i = 0; i < rows.size(); i++) {
                XWPFTableRow row = table.getRow(i);
                List<XWPFTableCell> cells = row.getTableCells();
                for (int j = 0; j < cells.size(); j++) {
                    XWPFTableCell cell = cells.get(j);
                    String key = cell.getText();
                    if (key.length() == 0) {
                        continue;
                    }
                    String keyHead = key.substring(0, 1);
                    String trueKey = key.substring(1);
                    if (key2methods.containsKey(keyHead)) {
                        cell.removeParagraph(0);
                    } else {
                        continue;
                    }
                    if (values.containsKey(trueKey)) {
                        runs.put(key, cell);
                    } else {
                        cell.setText("");
                    }
                }
            }
        }

        for (Entry<String, XWPFTableCell> entry : runs.entrySet()) {
            String key = entry.getKey();
            XWPFTableCell cell = entry.getValue();
            Method method = key2methods.get(key.subSequence(0, 1));
            String trueKey = key.substring(1);
            try {
                method.invoke(this, cell, values.get(trueKey));
            } catch (Exception e) {
                e.printStackTrace();
                cell.setText("");
            }
        }
        OutputStream outputStream = new FileOutputStream(outputPath);
        document.write(outputStream);
        outputStream.close();
        document.close();
        FileUtils.forceDelete(temp);
    }

    private void processText(XWPFTableCell cell, String msg) {
        // 执行\n换行
        if (msg == null) {
            cell.setText("");
            return;
        }
        String[] st = msg.split("\n");
        for (int p = 0; p < st.length; p++) {
            cell.addParagraph();
        }
        List<XWPFParagraph> pglist = cell.getParagraphs();
        for (int pi = 0; pi < pglist.size(); pi++) {
            XWPFRun run = pglist.get(pi).createRun();
            if (pi >= st.length) {
                System.out.println("ERROR!!!!!!!!!!!");
                continue;
            }
            run.setText(st[pi]);
        }

    }

    private void replaceInParagragh(XWPFParagraph paragraph, String oldText, String newText) {
        List<XWPFRun> runs = paragraph.getRuns();
        TextSegment tSegment = paragraph.searchText(oldText, new PositionInParagraph());
        if (tSegment != null) {
            int beginRun = tSegment.getBeginRun();
            int endRun = tSegment.getEndRun();
            if (beginRun == endRun) {
                XWPFRun run = runs.get(beginRun);
                String runText = run.getText(0);
                String replaced = runText.replace(oldText, newText);
                run.setText(replaced, 0);
            } else {
                StringBuilder b = new StringBuilder();
                for (int runPos = beginRun; runPos <= endRun; runPos++) {
                    XWPFRun run = runs.get(runPos);
                    b.append(run.getText(0));
                }
                String connectedRuns = b.toString();
                String replaced = connectedRuns.replace(oldText, newText);
                XWPFRun partOne = runs.get(beginRun);
                partOne.setText(replaced, 0);
                for (int runPos = beginRun + 1; runPos <= endRun; runPos++) {
                    XWPFRun partNext = runs.get(runPos);
                    partNext.setText("", 0);
                }
            }
        }
    }

    private void replaceInParagragh(XWPFParagraph paragraph, TextSegment tSegment, String oldText, String newText) {
        List<XWPFRun> runs = paragraph.getRuns();
        int beginRun = tSegment.getBeginRun();
        int endRun = tSegment.getEndRun();
        if (beginRun == endRun) {
            XWPFRun run = runs.get(beginRun);
            String runText = run.getText(0);
            String replaced = runText.replace(oldText, newText);
            int textIndex = replaced.indexOf(newText);
            // run.setText(replaced, 0);

            String fontFamily = run.getFontFamily();
            int fontSize = run.getFontSize();
            if (oldText.startsWith("□")) {
                XWPFRun replacedRun = null;

                String redundantText = replaced.substring(0, textIndex);
                if (redundantText.length() > 0) {
                    replacedRun = paragraph.insertNewRun(beginRun);
                    if (fontFamily != null) {
                        replacedRun.setFontFamily(fontFamily);
                    }
                    if (fontSize != -1) {
                        replacedRun.setFontSize(fontSize);
                    }
                    replacedRun.setText(redundantText, 0);
                    beginRun++;
                }

                replacedRun = paragraph.insertNewRun(beginRun);
                replacedRun.setFontFamily("MS Gothic");
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText("\u2611", 0);

                replacedRun = paragraph.insertNewRun(beginRun + 1);
                if (fontFamily != null) {
                    replacedRun.setFontFamily(fontFamily);
                }
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText(replaced.substring(textIndex), 0);
            } else {
                XWPFRun replacedRun = null;
                String redundantText = replaced.substring(0, textIndex + newText.length());
                replacedRun = paragraph.insertNewRun(beginRun);
                if (fontFamily != null) {
                    replacedRun.setFontFamily(fontFamily);
                }
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText(redundantText, 0);

                replacedRun = paragraph.insertNewRun(beginRun + 1);
                replacedRun.setFontFamily("MS Gothic");
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText("\u2611", 0);

                redundantText = replaced.substring(textIndex + newText.length());
                if (redundantText.length() > 0) {
                    replacedRun = paragraph.insertNewRun(beginRun + 2);
                    if (fontFamily != null) {
                        replacedRun.setFontFamily(fontFamily);
                    }
                    if (fontSize != -1) {
                        replacedRun.setFontSize(fontSize);
                    }
                    replacedRun.setText(redundantText, 0);
                    beginRun++;
                }
            }
        } else {
            StringBuilder b = new StringBuilder();
            for (int runPos = beginRun; runPos <= endRun; runPos++) {
                XWPFRun run = runs.get(runPos);
                b.append(run.getText(0));
            }
            String connectedRuns = b.toString();
            String replaced = connectedRuns.replace(oldText, newText);
            XWPFRun partOne = runs.get(beginRun);
            partOne.setText(replaced, 0);
            for (int runPos = beginRun + 1; runPos <= endRun; runPos++) {
                XWPFRun partNext = runs.get(runPos);
                partNext.setText("", 0);
            }

            int textIndex = replaced.indexOf(newText);
            String fontFamily = partOne.getFontFamily();
            int fontSize = partOne.getFontSize();
            if (oldText.startsWith("□")) {
                XWPFRun replacedRun = null;

                String redundantText = replaced.substring(0, textIndex);
                if (redundantText.length() > 0) {
                    replacedRun = paragraph.insertNewRun(beginRun);
                    if (fontFamily != null) {
                        replacedRun.setFontFamily(fontFamily);
                    }
                    if (fontSize != -1) {
                        replacedRun.setFontSize(fontSize);
                    }
                    replacedRun.setText(redundantText, 0);
                    beginRun++;
                }

                replacedRun = paragraph.insertNewRun(beginRun);
                replacedRun.setFontFamily("MS Gothic");
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText("\u2611", 0);

                replacedRun = paragraph.insertNewRun(beginRun + 1);
                if (fontFamily != null) {
                    replacedRun.setFontFamily(fontFamily);
                }
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText(replaced.substring(textIndex), 0);
            } else {
                XWPFRun replacedRun = null;
                String redundantText = replaced.substring(0, textIndex + newText.length());
                replacedRun = paragraph.insertNewRun(beginRun);
                if (fontFamily != null) {
                    replacedRun.setFontFamily(fontFamily);
                }
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText(redundantText, 0);

                replacedRun = paragraph.insertNewRun(beginRun + 1);
                replacedRun.setFontFamily("MS Gothic");
                if (fontSize != -1) {
                    replacedRun.setFontSize(fontSize);
                }
                replacedRun.setText("\u2611", 0);

                redundantText = replaced.substring(textIndex + newText.length());
                if (redundantText.length() > 0) {
                    replacedRun = paragraph.insertNewRun(beginRun + 2);
                    if (fontFamily != null) {
                        replacedRun.setFontFamily(fontFamily);
                    }
                    if (fontSize != -1) {
                        replacedRun.setFontSize(fontSize);
                    }
                    replacedRun.setText(redundantText, 0);
                    beginRun++;
                }
            }
        }
        paragraph.removeRun(beginRun + 2);
    }

    private void processImage(XWPFTableCell cell, String imgInfo) {
        String[] infos = imgInfo.split(":");
        String imgPath = infos[0];

        int imgHeight = Integer.parseInt(infos[1]);
        int imgWidth = Integer.parseInt(infos[2]);
        XWPFParagraph paragraph = cell.addParagraph();
        int length = paragraph.getRuns().size();
        if (length > 0) {
            for (int idx = (length - 1); idx >= 0; idx--) {
                paragraph.removeRun(idx);
            }
        }
        try {
            String blipId = document.addPictureData(new FileInputStream(new File(imgPath)), Document.PICTURE_TYPE_PNG);
            document.createPicture(blipId, document.getNextPicNameNumber(Document.PICTURE_TYPE_PNG), imgHeight,
                    imgWidth, paragraph);
        } catch (Exception e) {
            try {
                cell.removeParagraph(0);
                processText(cell, infos[3]);
            } catch (Exception e1) {
                cell.setText("");
                e1.printStackTrace();
            }
            // e.printStackTrace();
        }
    }

    private void processTable(XWPFTableCell cell, String tableInfos) {
        String[] infos = tableInfos.split("&");

        String tableMetaInfo = infos[0];
        String tableContent = infos[1];

        String[] mInfos = tableMetaInfo.split(":");
        int rows = Integer.parseInt(mInfos[0]);
        int cols = Integer.parseInt(mInfos[1]);

        XmlCursor cursor = cell.addParagraph().getCTP().newCursor();
        // cursor
        XWPFTable table = cell.insertNewTbl(cursor);

        String[] rowsString = tableContent.split("\r\n");

        for (int i = 0; i < rows; i++) {
            if (rowsString.length <= i) {
                break;
            }

            XWPFTableRow row = table.createRow();

            if (i == 0) {
                for (int j = 0; j < cols; j++) {
                    row.addNewTableCell();
                }
            }

            String[] colsString = rowsString[i].split("\t");

            for (int j = 0; j < cols; j++) {
                if (colsString.length <= j) {
                    break;
                }
                String msg = colsString[j];

                XWPFTableCell tableCell = row.getCell(j);
                tableCell.removeParagraph(0);
                try {
                    processText(tableCell, msg);
                } catch (Exception e) {
                    tableCell.setText("");
                }

            }
        }

    }

    public static String generateImageStr(String path, int w, int h, String alt) {
        return path + ":" + String.valueOf(w) + ":" + String.valueOf(h) + ":" + alt;
    }

}
